"""
Module for Proper Orthogonal Decomposition (POD).
Three different methods can be employed: Truncated Singular Value
Decomposition, Truncated Randomized Singular Value Decomposition, Truncated
Singular Value Decomposition via correlation matrix.
"""
try:
    from scipy.linalg import eigh
except ImportError:
    from numpy.linalg import eigh
import numpy as np

from .reduction import Reduction


class POD(Reduction):
    def __init__(self, method='svd', **kwargs):
        """
        Perform the Proper Orthogonal Decomposition.

        :param method: the implementation to use for the computation of the POD
            modes. Default is 'svd'.
        :type method: {'svd', 'randomized_svd', 'correlation_matrix'}
        :param rank: the rank for the truncation; If 0, the method computes
            the optimal rank and uses it for truncation; if positive interger,
            the method uses the argument for the truncation; if float between 0
            and 1, the rank is the number of the biggest singular values that
            are needed to reach the 'energy' specified by `svd_rank`; if -1,
            the method does not compute truncation. Default is 0. The `rank`
            parameter is available using all the available methods.
        :type rank: int or float
        :param int subspace_iteration: the number of subspace iteration in the
            randomized svd. It is available only using the 'randomized_svd'
            method. Default value is 1.
        :param int omega_rank: the number of columns of the Omega random
            matrix. If set to 0, the number of columns is equal to twice the
            `rank` (if it has explicitly passed as integer) or twice the number
            of input snapshots. Default value is 0. It is available only using
            the 'randomized_svd' method.
        :param bool save_memory: reduce the usage of the memory, despite an
            higher number of operations. It is available only using the
            'correlation_matrix' method. Default value is False.


        :Example:
            >>> pod = POD().fit(snapshots)
            >>> reduced_snapshots = pod.reduce(snapshots)
            >>> # Other possible constructors are ...
            >>> pod = POD('svd')
            >>> pod = POD('svd', rank=20)
            >>> pod = POD('randomized_svd', rank=-1)
            >>> pod = POD('randomized_svd', rank=0, subspace_iteration=3,
                          omega_rank=10)
            >>> pod = POD('correlation_matrix', rank=10, save_memory=False)
        """
        available_methods = {
            'svd': (self._svd, {
                'rank': -1
            }),
            'randomized_svd': (self._rsvd, {
                'rank': -1,
                'subspace_iteration': 1,
                'omega_rank': 0
            }),
            'correlation_matrix': (self._corrm, {
                'rank': -1,
                'save_memory': False
            }),
        }

        self._modes = None
        self._singular_values = None

        method = available_methods.get(method)
        if method is None:
            raise RuntimeError(
                "Invalid method for POD. Please chose one among {}".format(
                    ', '.join(available_methods)))

        self.__method, args = method
        args.update(kwargs)

        for hyperparam, value in args.items():
            setattr(self, hyperparam, value)

    @property
    def modes(self):
        """
        The POD modes.

        :type: numpy.ndarray
        """
        return self._modes

    @property
    def singular_values(self):
        """
        The singular values

        :type: numpy.ndarray
        """
        return self._singular_values

    def fit(self, X):
        """
        Create the reduced space for the given snapshots `X` using the
        specified method

        :param numpy.ndarray X: the input snapshots matrix (stored by column)
        """
        self._modes, self._singular_values = self.__method(X)
        return self

    def transform(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).
        """
        return self.modes.T.conj().dot(X)

    def inverse_transform(self, X):
        """
        Projects a reduced to full order solution.

        :type: numpy.ndarray
        """
        return self.modes.dot(X)

    def reduce(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).

        .. note::

            Same as `transform`. Kept for backward compatibility.
        """
        return self.transform(X)

    def expand(self, X):
        """
        Projects a reduced to full order solution.

        :type: numpy.ndarray

        .. note::

            Same as `inverse_transform`. Kept for backward compatibility.
        """
        return self.inverse_transform(X)

    def _truncation(self, X, s):
        """
        Return the number of modes to select according to the `rank` parameter.
        See POD.__init__ for further info.

        :param numpy.ndarray X: the matrix to decompose.
        :param numpy.ndarray s: the singular values of X.

        :return: the number of modes
        :rtype: int
        """
        def omega(x):
            return 0.56 * x**3 - 0.95 * x**2 + 1.82 * x + 1.43

        if self.rank == 0:
            beta = np.divide(*sorted(X.shape))
            tau = np.median(s) * omega(beta)
            rank = np.sum(s > tau)
        elif self.rank > 0 and self.rank < 1:
            cumulative_energy = np.cumsum(s**2 / (s**2).sum())
            rank = np.searchsorted(cumulative_energy, self.rank) + 1
        elif self.rank >= 1 and isinstance(self.rank, int):
            rank = self.rank
        else:
            rank = X.shape[1]

        return rank

    def _svd(self, X):
        """
        Truncated Singular Value Decomposition.

        :param numpy.ndarray X: the matrix to decompose.
        :return: the truncated left-singular vectors matrix, the truncated
            singular values array, the truncated right-singular vectors matrix.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray

        References:
        Gavish, Matan, and David L. Donoho, The optimal hard threshold for
        singular values is, IEEE Transactions on Information Theory 60.8
        (2014): 5040-5053.
        """
        U, s = np.linalg.svd(X, full_matrices=False)[:2]

        rank = self._truncation(X, s)
        return U[:, :rank], s[:rank]

    def _rsvd(self, X):
        """
        Truncated randomized Singular Value Decomposition.

        :param numpy.ndarray X: the matrix to decompose.
        :return: the truncated left-singular vectors matrix, the truncated
            singular values array, the truncated right-singular vectors matrix.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray

        References:
        Finding structure with randomness: probabilistic algorithms for
        constructing approximate matrix decompositions. N. Halko, P. G.
        Martinsson, J. A. Tropp.
        """
        if (self.omega_rank == 0 and isinstance(self.rank, int)
                and self.rank not in [0, -1]):
            omega_rank = self.rank * 2
        elif self.omega_rank == 0:
            omega_rank = X.shape[1] * 2
        else:
            omega_rank = self.omega_rank
        Omega = np.random.rand(X.shape[1], omega_rank)

        Y = np.dot(X, Omega)
        Q = np.linalg.qr(Y)[0]

        if self.subspace_iteration:
            for _ in range(self.subspace_iteration):
                Y_ = np.dot(X.T.conj(), Q)
                Q_ = np.linalg.qr(Y_)[0]
                Y = np.dot(X, Q_)
                Q = np.linalg.qr(Y)[0]

        B = np.dot(Q.T.conj(), X)
        Uy, s = np.linalg.svd(B, full_matrices=False)[:2]
        U = Q.dot(Uy)

        rank = self._truncation(X, s)
        return U[:, :rank], s[:rank]

    def _corrm(self, X):
        """
        Truncated POD calculated with correlation matrix.

        :param numpy.ndarray X: the matrix to decompose.
        :return: the truncated left-singular vectors matrix, the truncated
            singular values array, the truncated right-singular vectors matrix.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray
        """

        if self.save_memory:
            corr = np.empty(shape=(X.shape[1], X.shape[1]))
            for i, i_snap in enumerate(X.T):
                for j, k_snap in enumerate(X.T):
                    corr[i, j] = np.inner(i_snap, k_snap)

        else:
            corr = X.T.dot(X)

        eigs, eigv = eigh(corr)

        ordered_idx = np.argsort(eigs)[::-1]
        eigs = eigs[ordered_idx]
        eigv = eigv[:, ordered_idx]
        s = np.sqrt(eigs[eigs > 0])
        rank = self._truncation(X, s)

        # compute modes
        eigv = eigv[:, eigs > 0]
        U = X.dot(eigv) / s

        return U[:, :rank], s[:rank]



"""Wrapper for RadiusNeighborsRegressor."""

from sklearn.neighbors import RadiusNeighborsRegressor as Regressor

from .neighbors_regressor import NeighborsRegressor


class RadiusNeighborsRegressor(NeighborsRegressor):
    """
    Radius Neighbors Regressor.

    :param kwargs: arguments passed to the internal instance of
        RadiusNeighborsRegressor.
    """
    def __init__(self, **kwargs):
        self.regressor = Regressor(**kwargs)



"""EZyRB package"""

__all__ = [
    'Database', 'Reduction', 'POD', 'Approximation', 'RBF', 'Linear', 'GPR',
    'ANN', 'KNeighborsRegressor', 'RadiusNeighborsRegressor', 'AE',
    'ReducedOrderModel', 'PODAE'
]

from .meta import *
from .database import Database
from .reduction import Reduction
from .pod import POD
from .ae import AE
from .pod_ae import PODAE
from .approximation import Approximation
from .rbf import RBF
from .linear import Linear
from .gpr import GPR
from .reducedordermodel import ReducedOrderModel
from .ann import ANN
from .kneighbors_regressor import KNeighborsRegressor
from .radius_neighbors_regressor import RadiusNeighborsRegressor



"""Wrapper for K-Neighbors Regressor."""

from sklearn.neighbors import KNeighborsRegressor as Regressor

from .neighbors_regressor import NeighborsRegressor


class KNeighborsRegressor(NeighborsRegressor):
    """
    K-Neighbors Regressor.

    :param kwargs: arguments passed to the internal instance of
        KNeighborsRegressor.
    """
    def __init__(self, **kwargs):
        self.regressor = Regressor(**kwargs)



"""
Module for Linear N-Dimensional Interpolation
"""

import numpy as np
from scipy.interpolate import LinearNDInterpolator as LinearNDInterp
from scipy.interpolate import interp1d

from .approximation import Approximation


class Linear(Approximation):
    """
    Multidimensional linear interpolator.

    :param float fill_value: value used to fill in for requested points outside
        of the convex hull of the input points. If not provided, then the
        default is numpy.nan.
    """
    def __init__(self, fill_value=np.nan):
        self.fill_value = fill_value
        self.interpolator = None

    def fit(self, points, values):
        """
        Construct the interpolator given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        # the first dimension is the list of parameters, the second one is
        # the dimensionality of each tuple of parameters (we look for
        # parameters of dimensionality one)
        as_np_array = np.array(points)
        if not np.issubdtype(as_np_array.dtype, np.number):
            raise ValueError('Invalid format or dimension for the argument'
                             '`points`.')

        if as_np_array.shape[-1] == 1:
            as_np_array = np.squeeze(as_np_array, axis=-1)

        if as_np_array.ndim == 1 or (as_np_array.ndim == 2
                                     and as_np_array.shape[1] == 1):
            self.interpolator = interp1d(as_np_array, values, axis=0)
        else:
            self.interpolator = LinearNDInterp(points,
                                               values,
                                               fill_value=self.fill_value)

    def predict(self, new_point):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        return self.interpolator(new_point)



"""Module for the snapshots database collected during the Offline stage."""

import numpy as np

class Database():
    """
    Database class

    :param array_like parameters: the input parameters
    :param array_like snapshots: the input snapshots
    :param Scale scaler_parameters: the scaler for the parameters. Default
        is None meaning no scaling.
    :param Scale scaler_snapshots: the scaler for the snapshots. Default is
        None meaning no scaling.
    :param array_like space: the input spatial data
    """
    def __init__(self,
                 parameters=None,
                 snapshots=None,
                 scaler_parameters=None,
                 scaler_snapshots=None,
                 space=None):
        self._parameters = None
        self._snapshots = None
        self._space = None
        self.scaler_parameters = scaler_parameters
        self.scaler_snapshots = scaler_snapshots

        # if only parameters or snapshots are provided
        if (parameters is None) ^ (snapshots is None):
            raise RuntimeError(
                'Parameters and Snapshots are not both provided')

        if space is not None and snapshots is None:
            raise RuntimeError(
                'Snapshot data is not provided with Spatial data')

        if space is not None and snapshots is None:
            raise RuntimeError

        if space is not None and snapshots is None:
            raise RuntimeError(
                'Snapshot data is not provided with Spatial data')

        if parameters is not None and snapshots is not None:
            if space is not None:
                self.add(parameters, snapshots, space)
            else:
                self.add(parameters, snapshots)


    @property
    def parameters(self):
        """
        The matrix containing the input parameters (by row).

        :rtype: numpy.ndarray
        """
        if self.scaler_parameters:
            return self.scaler_parameters.fit_transform(self._parameters)

        return self._parameters

    @property
    def snapshots(self):
        """
        The matrix containing the snapshots (by row).

        :rtype: numpy.ndarray
        """
        if self.scaler_snapshots:
            return self.scaler_snapshots.fit_transform(self._snapshots)

        return self._snapshots

    @property
    def space(self):
        """
        The matrix containing spatial information (by row).

        :rtype: numpy.ndarray
        """
        return self._space

    def __getitem__(self, val):
        """
        This method returns a new Database with the selected parameters and
        snapshots.

        .. warning:: The new parameters and snapshots are a view of the
            original Database.
        """
        if isinstance(val, int):
            if self._space is None:
                return Database(np.reshape(self._parameters[val],
                                    (1,len(self._parameters[val]))),
                                np.reshape(self._snapshots[val],
                                    (1,len(self._snapshots[val]))),
                                self.scaler_parameters,
                                self.scaler_snapshots)

            return Database(np.reshape(self._parameters[val],
                                (1,len(self._parameters[val]))),
                            np.reshape(self._snapshots[val],
                                (1,len(self._snapshots[val]))),
                            self.scaler_parameters,
                            self.scaler_snapshots,
                            np.reshape(self._space[val],
                                (1,len(self._space[val]))))

        if self._space is None:
            return Database(self._parameters[val],
                            self._snapshots[val],
                            self.scaler_parameters,
                            self.scaler_snapshots)

        return Database(self._parameters[val],
                        self._snapshots[val],
                        self.scaler_parameters,
                        self.scaler_snapshots,
                        self._space[val])

    def __len__(self):
        """
        This method returns the number of snapshots.

        :rtype: int
        """
        return len(self._snapshots)

    def add(self, parameters, snapshots, space=None):
        """
        Add (by row) new sets of snapshots and parameters to the original
        database.

        :param array_like parameters: the parameters to add.
        :param array_like snapshots: the snapshots to add.
        """
        if len(parameters) != len(snapshots):
            raise RuntimeError(
                'Different number of parameters and snapshots.')

        if self._space is not None:
            if space is None:
                raise RuntimeError('No Spatial Value given')

        if (self._space is not None) or (space is not None):
            if space.shape != snapshots.shape:
                raise RuntimeError(
                    'shape of space and snapshots are different.')

        if self._parameters is None and self._snapshots is None:
            self._parameters = parameters
            self._snapshots = snapshots
            if self._space is None:
                self._space = space
        elif self._space is None:
            self._parameters = np.vstack([self._parameters, parameters])
            self._snapshots = np.vstack([self._snapshots, snapshots])
        else:
            self._parameters = np.vstack([self._parameters, parameters])
            self._snapshots = np.vstack([self._snapshots, snapshots])
            self._space = np.vstack([self._space, space])

        return self



"""Module for generic NeighborsRegressor."""

import numpy as np
from .approximation import Approximation


class NeighborsRegressor(Approximation):
    """
    A generic superclass for wrappers of *NeighborsRegressor from sklearn.

    :param kwargs: arguments passed to the internal instance of
        *NeighborsRegressor.
    """
    def fit(self, points, values):
        """
        Construct the interpolator given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        points = np.array(points).reshape(len(points), -1)
        values = np.array(values)

        self.regressor.fit(points, values)

    def predict(self, new_point):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        if isinstance(new_point, (list, np.ndarray)):
            new_point = np.array(new_point).reshape(len(new_point), -1)
        else:
            new_point = np.array([new_point])

        return self.regressor.predict(new_point)



"""Module for the Reduction abstract class."""

from abc import ABC, abstractmethod


class Reduction(ABC):
    """
    The abstract `Approximation` class.

    All the classes that implement the input-output mapping should be inherited
    from this class.
    """
    @abstractmethod
    def fit(self):
        """Abstract `fit`"""

    @abstractmethod
    def transform(self):
        """Abstract `transform`"""

    @abstractmethod
    def inverse_transform(self):
        """Abstract `inverse_transform`"""



"""
Module for FNN-Autoencoders.
"""

import torch
import torch.nn as nn
from .reduction import Reduction
from .ae import AE
from .pod import POD


class PODAE(POD, AE):
    """
    Feed-Forward AutoEncoder class with POD (AE)
    """
    def __init__(self, pod, ae):
        self.pod = pod
        self.ae = ae

    def fit(self, X):
        """
        """
        self.pod.fit(X)
        coefficients = self.pod.transform(X)
        self.ae.fit(coefficients)

    def transform(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).
        """
        coeff = self.pod.reduce(X)
        coeff = self._convert_numpy_to_torch(coeff).T
        g = self.ae.encoder(coeff)
        return g.cpu().detach().numpy().T

    def inverse_transform(self, g):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.
        """
        g = self._convert_numpy_to_torch(g).T
        u = self.ae.decoder(g)
        u = u.cpu().detach().numpy().T
        return self.pod.expand(u)

    def reduce(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).

        .. note::

            Same as `transform`. Kept for backward compatibility.
        """
        return self.transform(X)

    def expand(self, g):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.
        
        .. note::

            Same as `inverse_transform`. Kept for backward compatibility.
        """
        return self.inverse_transform(g)



"""Module for the Approximation abstract class"""

from abc import ABC, abstractmethod


class Approximation(ABC):
    """
    The abstract `Approximation` class.

    All the classes that implement the input-output mapping should be inherited
    from this class.
    """
    @abstractmethod
    def fit(self, points, values):
        """Abstract `fit`"""

    @abstractmethod
    def predict(self, new_point):
        """Abstract `predict`"""



"""
Module for FNN-Autoencoders.
"""

import torch
import torch.nn as nn
from .reduction import Reduction
from .ann import ANN


class AE(Reduction, ANN):
    """
    Feed-Forward AutoEncoder class (AE)

    :param list layers_encoder: ordered list with the number of neurons of
        each hidden layer for the encoder
    :param list layers_decoder: ordered list with the number of neurons of
        each hidden layer for the decoder
    :param torch.nn.modules.activation function_encoder: activation function
        at each layer for the encoder, except for the output layer at with
        Identity is considered by default.  A single activation function can
        be passed or a list of them of length equal to the number of hidden
        layers.
    :param torch.nn.modules.activation function_decoder: activation function
        at each layer for the decoder, except for the output layer at with
        Identity is considered by default.  A single activation function can
        be passed or a list of them of length equal to the number of hidden
        layers.
    :param list stop_training: list with the maximum number of training
        iterations (int) and/or the desired tolerance on the training loss
        (float).
    :param torch.nn.Module loss: loss definition (Mean Squared if not
        given).
    :param torch.optim optimizer: the torch class implementing optimizer.
        Default value is `Adam` optimizer.
    :param float lr: the learning rate. Default is 0.001.

    :Example:
        >>> from ezyrb import AE
        >>> import torch
        >>> f = torch.nn.Softplus
        >>> low_dim = 5
        >>> optim = torch.optim.Adam
        >>> ae = AE([400, low_dim], [low_dim, 400], f(), f(), 2000)
        >>> # or ...
        >>> ae = AE([400, 10, 10, low_dim], [low_dim, 400], f(), f(), 1e-5,
        >>>          optimizer=optim)
        >>> ae.fit(snapshots)
        >>> reduced_snapshots = ae.reduce(snapshots)
        >>> expanded_snapshots = ae.expand(reduced_snapshots)
    """
    def __init__(self,
                 layers_encoder,
                 layers_decoder,
                 function_encoder,
                 function_decoder,
                 stop_training,
                 loss=None,
                 optimizer=torch.optim.Adam,
                 lr=0.001):

        if loss is None:
            loss = torch.nn.MSELoss()

        if not isinstance(function_encoder, list):
            # Single activation function passed
            function_encoder = [function_encoder] * (len(layers_encoder))
        if not isinstance(function_decoder, list):
            # Single activation function passed
            function_decoder = [function_decoder] * (len(layers_decoder))

        if not isinstance(stop_training, list):
            stop_training = [stop_training]

        self.layers_encoder = layers_encoder
        self.layers_decoder = layers_decoder
        self.function_encoder = function_encoder
        self.function_decoder = function_decoder
        self.loss = loss

        self.stop_training = stop_training
        self.loss_trend = []
        self.encoder = None
        self.decoder = None
        self.optimizer = optimizer
        self.model = None
        self.lr = lr

    class InnerAE(torch.nn.Module):
        """
        Autoencoder as a pytorch module
        """
        def __init__(self, outer_instance):
            super().__init__()
            self.encoder = outer_instance.encoder
            self.decoder = outer_instance.decoder

        def forward(self, x):
            """
            Compute the forward of the autoencoder
            """
            encoded = self.encoder(x)
            decoded = self.decoder(encoded)
            return decoded

    def _build_model(self, values):
        """
        Build the torch model.

        Considering the number of neurons per layer (self.layers), a
        feed-forward NN is defined:
            - activation function from layer i>=0 to layer i+1:
              self.function[i]; activation function at the output layer:
              Identity (by default).

        :param numpy.ndarray values: the set values one wants to reduce.
        """
        layers_encoder = self.layers_encoder.copy()
        layers_encoder.insert(0, values.shape[1])

        layers_decoder = self.layers_decoder.copy()
        layers_decoder.append(values.shape[1])

        layers_encoder_torch = []
        for i in range(len(layers_encoder) - 2):
            layers_encoder_torch.append(
                nn.Linear(layers_encoder[i], layers_encoder[i + 1]))
            layers_encoder_torch.append(self.function_encoder[i])
        layers_encoder_torch.append(
            nn.Linear(layers_encoder[-2], layers_encoder[-1]))

        layers_decoder_torch = []
        for i in range(len(layers_decoder) - 2):
            layers_decoder_torch.append(
                nn.Linear(layers_decoder[i], layers_decoder[i + 1]))
            layers_decoder_torch.append(self.function_decoder[i])
        layers_decoder_torch.append(
            nn.Linear(layers_decoder[-2], layers_decoder[-1]))
        self.encoder = nn.Sequential(*layers_encoder_torch)
        self.decoder = nn.Sequential(*layers_decoder_torch)
        # Creating the model adding the encoder and the decoder
        self.model = self.InnerAE(self)

    def fit(self, values):
        """
        Build the AE given 'values' and perform training.

        Training procedure information:
            -  optimizer: Adam's method with default parameters (see, e.g.,
               https://pytorch.org/docs/stable/optim.html);
            -  loss: self.loss (if none, the Mean Squared Loss is set by
               default).
            -  stopping criterion: the fulfillment of the requested tolerance
               on the training loss compatibly with the prescribed budget of
               training iterations (if type(self.stop_training) is list); if
               type(self.stop_training) is int or type(self.stop_training) is
               float, only the number of maximum iterations or the accuracy
               level on the training loss is considered as the stopping rule,
               respectively.

        :param numpy.ndarray values: the (training) values in the points.
        """
        values = values.T
        self._build_model(values)
        self.optimizer = self.optimizer(self.model.parameters(), lr=self.lr)
        values = self._convert_numpy_to_torch(values)
        n_epoch = 1
        flag = True
        while flag:
            y_pred = self.model(values)
            loss = self.loss(y_pred, values)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.loss_trend.append(loss.item())
            for criteria in self.stop_training:
                if isinstance(criteria, int):
                    # stop criteria is an integer
                    if n_epoch == criteria:
                        flag = False
                elif isinstance(criteria, float):
                    # stop criteria is float
                    if loss.item() < criteria:
                        flag = False

            n_epoch += 1

    def transform(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).
        """
        X = self._convert_numpy_to_torch(X).T
        g = self.encoder(X)
        return g.cpu().detach().numpy().T

    def inverse_transform(self, g):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.
        """
        g = self._convert_numpy_to_torch(g).T
        u = self.decoder(g)
        return u.cpu().detach().numpy().T

    def reduce(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).

        .. note::

            Same as `transform`. Kept for backward compatibility.
        """
        return self.transform(X)

    def expand(self, g):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.

        .. note::

            Same as `inverse_transform`. Kept for backward compatibility.
        """
        return self.inverse_transform(g)



"""
Module wrapper exploiting `GPy` for Gaussian Process Regression
"""
import numpy as np
from scipy.optimize import minimize
from sklearn.gaussian_process import GaussianProcessRegressor

from .approximation import Approximation


class GPR(Approximation):
    """
    Multidimensional regression using Gaussian process.

    :cvar numpy.ndarray X_sample: the array containing the input points,
        arranged by row.
    :cvar numpy.ndarray Y_sample: the array containing the output values,
        arranged by row.
    :cvar GPy.models.GPRegression model: the regression model.

    :Example:

        >>> import ezyrb
        >>> import numpy as np
        >>> x = np.random.uniform(-1, 1, size=(4, 2))
        >>> y = (np.sin(x[:, 0]) + np.cos(x[:, 1]**3)).reshape(-1, 1)
        >>> gpr = ezyrb.GPR()
        >>> gpr.fit(x, y)
        >>> y_pred = gpr.predict(x)
        >>> print(np.allclose(y, y_pred))

    """
    def __init__(self):
        self.X_sample = None
        self.Y_sample = None
        self.model = None

    def fit(self,
            points,
            values,
            kern=None,
            normalizer=True,
            optimization_restart=20):
        """
        Construct the regression given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        :param sklearn.gaussian_process.kernels.Kernel kern: kernel object from
            sklearn.
        :param bool normalizer: whether to normilize `values` or not.
            Defaults to True.
        :param int optimization_restart: number of restarts for the
            optimization. Defaults to 20.
        """
        self.X_sample = np.array(points)
        self.Y_sample = np.array(values)
        if self.X_sample.ndim == 1:
            self.X_sample = self.X_sample.reshape(-1, 1)
        if self.Y_sample.ndim == 1:
            self.Y_sample = self.Y_sample.reshape(-1, 1)

        self.model = GaussianProcessRegressor(
            kernel=kern, n_restarts_optimizer=optimization_restart,
            normalize_y=normalizer)
        self.model.fit(self.X_sample, self.Y_sample)

    def predict(self, new_points, return_variance=False):
        """
        Predict the mean and the variance of Gaussian distribution at given
        `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :param bool return_variance: flag to return also the variance.
            Default is False.
        :return: the mean and the variance
        :rtype: (numpy.ndarray, numpy.ndarray)
        """
        new_points = np.atleast_2d(new_points)
        return self.model.predict(new_points, return_std=return_variance)

    def optimal_mu(self, bounds, optimization_restart=10):
        """
        Proposes the next sampling point by looking at the point where the
        Gaussian covariance is maximized. A gradient method (with multi
        starting points) is adopted for the optimization.

        :param numpy.ndarray bounds: the boundaries in the gradient
            optimization. The shape must be (*input_dim*, 2), where *input_dim*
            is the dimension of the input points.
        :param int optimization_restart: the number of restart in the gradient
            optimization. Default is 10.
        """
        dim = self.X_sample.shape[1]
        min_val = 1
        min_x = None

        def min_obj(X):
            return -1 * np.linalg.norm(self.predict(X.reshape(1, -1), True)[1])

        initial_starts = np.random.uniform(bounds[:, 0],
                                           bounds[:, 1],
                                           size=(optimization_restart, dim))

        # Find the best optimum by starting from n_restart different random
        # points.
        for x0 in initial_starts:
            res = minimize(min_obj, x0, bounds=bounds, method='L-BFGS-B')

            if res.fun < min_val:
                min_val = res.fun
                min_x = res.x

        return min_x.reshape(1, -1)



"""EZyRB metadata."""

__all__ = [
    '__project__', '__title__', '__author__', '__copyright__', '__license__',
    '__version__', '__mail__', '__maintainer__', '__status__'
]

__project__ = 'EZyRB'
__title__ = "ezyrb"
__author__ = "Nicola Demo, Marco Tezzele"
__copyright__ = "Copyright 2016-2021, EZyRB contributors"
__license__ = "MIT"
__version__ = "1.2.1.post2207"
__mail__ = 'demo.nicola@gmail.com, marcotez@gmail.com'
__maintainer__ = __author__
__status__ = "Stable"



"""Module for Radial Basis Function Interpolation."""

import numpy as np
from scipy.interpolate import RBFInterpolator
from .approximation import Approximation


class RBF(Approximation):
    """
    Multidimensional interpolator using Radial Basis Function.

    :param kernel: The radial basis function; the default is ‘multiquadric’.
    :type kernel: str or callable
    :param float smooth: values greater than zero increase the smoothness of
        the approximation. 0 is for interpolation (default), the function will
        always go through the nodal points in this case.
    :param int neighbors: if specified, the value of the interpolant at each
        evaluation point will be computed using only the nearest data points.
        If None (default), all the data points are used by default.
    :param float epsilon: Shape parameter that scales the input to the RBF.
        If kernel is ‘linear’, ‘thin_plate_spline’, ‘cubic’, or ‘quintic’, this
        defaults to 1 and can be ignored. Otherwise, this must be specified.
    :param int degree: Degree of the added polynomial. The default value is
        the minimum degree for kernel or 0 if there is no minimum degree.

    :cvar kernel: The radial basis function; the default is ‘multiquadric’.
    :cvar interpolator: the RBF interpolator

    :Example:

         >>> import ezyrb
         >>> import numpy as np
         >>>
         >>> x = np.random.uniform(-1, 1, size=(4, 2))
         >>> y = np.array([np.sin(x[:, 0]), np.cos(x[:, 1]**3)]).T
         >>> rbf = ezyrb.RBF()
         >>> rbf.fit(x, y)
         >>> y_pred = rbf.predict(x)
         >>> print(np.allclose(y, y_pred))

    """
    def __init__(self,
                 kernel='thin_plate_spline',
                 smooth=0,
                 neighbors=None,
                 epsilon=None,
                 degree=None):
        self.kernel = kernel
        self.smooth = smooth
        self.neighbors = neighbors
        self.degree = degree
        self.epsilon = epsilon
        self.interpolator = None
        self.xi = None

    def fit(self, points, values):
        """
        Construct the interpolator given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        self.xi = np.asarray(points)
        N = self.xi.shape[-1]

        if self.epsilon is None:
            # default epsilon is the "the average distance between nodes" based
            # on a bounding hypercube
            ximax = np.amax(self.xi, axis=0)
            ximin = np.amin(self.xi, axis=0)
            edges = ximax - ximin
            edges = edges[np.nonzero(edges)]
            self.epsilon = np.power(np.prod(edges)/N, 1.0/edges.size)
            if self.kernel in ['thin_plate_spline', 'cubic', 'quintic']:
                self.epsilon = 1

        self.interpolator = RBFInterpolator(
            points,
            values,
            neighbors=self.neighbors,
            smoothing=self.smooth,
            kernel=self.kernel,
            epsilon=self.epsilon,
            degree=self.degree)

    def predict(self, new_point):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        return self.interpolator(np.asarray(new_point))



from setuptools import setup

meta = {}
with open("ezyrb/meta.py") as fp:
    exec(fp.read(), meta)

# Package meta-data.
NAME = meta['__title__']
DESCRIPTION = 'Easy Reduced Basis'
URL = 'https://github.com/mathLab/EZyRB'
MAIL = meta['__mail__']
AUTHOR = meta['__author__']
VERSION = meta['__version__']
KEYWORDS = 'pod interpolation reduced-basis model-order-reduction'

REQUIRED = [
    'future', 'numpy', 'scipy',	'matplotlib', 'scikit-learn>=1.0', 'torch'
]

EXTRAS = {
    'docs': ['sphinx', 'sphinx_rtd_theme'],
    'test': ['pytest', 'pytest-cov'],
}

LDESCRIPTION = (
    "EZyRB is a Python package that performs a data-driven model order "
    "reduction for parametrized problems exploiting the recent approaches. "
    "Such techniques are able to provide a parametric model capable to "
    "provide the real-time approximation of the solution of a generic "
    "(potentially complex and non linear) problem. The reduced model is "
    "totally built upon the numerical data obtained by the original (to "
    "reduce) model, without requiring any knowledge of the equations that "
    "describe this model, resulting in a well suited framework for industrial "
    "contexts due to its natural capability to been integrated with "
    "commercial software."
)

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LDESCRIPTION,
    author=AUTHOR,
    author_email=MAIL,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    keywords=KEYWORDS,
    url=URL,
    license='MIT',
    packages=[NAME],
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    zip_safe=False,
)



"""
Module for Artificial Neural Network (ANN) Prediction.
"""

import torch
import torch.nn as nn
import numpy as np
from .approximation import Approximation


class ANN(Approximation):
    def __init__(self, layers, function, stop_training, loss=None):
        """
        Feed-Forward Artifical Neural Network (ANN).

        :param list layers: ordered list with the number of neurons of each
            hidden layer.
        :param torch.nn.modules.activation function: activation function at each
            layer, except for the output layer at with Identity is considered by
            default. A single activation function can be passed or a list of
            them of length equal to the number of hidden layers.
        :param list stop_training: list with the maximum number of training
            iterations (int) and/or the desired tolerance on the training loss
            (float).
        :param torch.nn.Module loss: loss definition (Mean Squared if not
            given).

        :Example:
            >>> import ezyrb
            >>> import numpy as np
            >>> import torch.nn as nn
            >>> x = np.random.uniform(-1, 1, size =(4, 2))
            >>> y = np.array([np.sin(x[:, 0]), np.cos(x[:, 1]**3)]).T
            >>> ann = ezyrb.ANN([10, 5], nn.Tanh(), [20000,1e-5])
            >>> ann.fit(x, y)
            >>> y_pred = ann.predict(x)
            >>> print(y)
            >>> print(y_pred)
            >>> print(len(ann.loss_trend))
            >>> print(ann.loss_trend[-1])
        """
        if loss is None:
            loss = torch.nn.MSELoss()

        if not isinstance(function, list):  # Single activation function passed
            function = [function] * (len(layers))

        if not isinstance(stop_training, list):
            stop_training = [stop_training]

        self.layers = layers
        self.function = function
        self.loss = loss
        self.stop_training = stop_training

        self.loss_trend = []
        self.model = None
        self.optimizer = None

    def _convert_numpy_to_torch(self, array):
        """
        Converting data type.

        :param numpy.ndarray array: input array.
        :return: the tensorial counter-part of the input array.
        :rtype: torch.Tensor.
        """
        return torch.from_numpy(array).float()

    def _convert_torch_to_numpy(self, tensor):
        """
        Converting data type.

        :param torch.Tensor tensor: input tensor.
        :return: the vectorial counter-part of the input tensor.
        :rtype: numpy.ndarray.
        """
        return tensor.detach().numpy()

    def _build_model(self, points, values):
        """
        Build the torch model.
        Considering the number of neurons per layer (self.layers), a
        feed-forward NN is defined:
        -  activation function from layer i>=0 to layer i+1:
           self.function[i]; activation function at the output layer:
           Identity (by default).
        :param numpy.ndarray points: the coordinates of the given (training)
        points.
        :param numpy.ndarray values: the (training) values in the points.
        """
        layers = self.layers.copy()
        layers.insert(0, points.shape[1])
        layers.append(values.shape[1])
        layers_torch = []
        for i in range(len(layers) - 2):
            layers_torch.append(nn.Linear(layers[i], layers[i + 1]))
            layers_torch.append(self.function[i])
        layers_torch.append(nn.Linear(layers[-2], layers[-1]))
        self.model = nn.Sequential(*layers_torch)

    def fit(self, points, values):
        """
        Build the ANN given 'points' and 'values' and perform training.

        Training procedure information:
            -  optimizer: Adam's method with default parameters (see, e.g.,
               https://pytorch.org/docs/stable/optim.html);
            -  loss: self.loss (if none, the Mean Squared Loss is set by
               default).
            -  stopping criterion: the fulfillment of the requested tolerance on
               the training loss compatibly with the prescribed budget of
               training iterations (if type(self.stop_training) is list); if
               type(self.stop_training) is int or type(self.stop_training) is
               float, only the number of maximum iterations or the accuracy
               level on the training loss is considered as the stopping rule,
               respectively.

        :param numpy.ndarray points: the coordinates of the given (training)
            points.
        :param numpy.ndarray values: the (training) values in the points.
        """

        self._build_model(points, values)
        self.optimizer = torch.optim.Adam(self.model.parameters())

        points = self._convert_numpy_to_torch(points)
        values = self._convert_numpy_to_torch(values)

        n_epoch = 1
        flag = True
        while flag:
            y_pred = self.model(points)
            loss = self.loss(y_pred, values)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.loss_trend.append(loss.item())
            for criteria in self.stop_training:
                if isinstance(criteria, int):  # stop criteria is an integer
                    if n_epoch == criteria:
                        flag = False
                elif isinstance(criteria, float):  # stop criteria is float
                    if loss.item() < criteria:
                        flag = False

            n_epoch += 1

    def predict(self, new_point):
        """
        Evaluate the ANN at given 'new_points'.

        :param array_like new_points: the coordinates of the given points.
        :return: the predicted values via the ANN.
        :rtype: numpy.ndarray
        """
        new_point = self._convert_numpy_to_torch(np.array(new_point))
        y_new = self.model(new_point)
        return self._convert_torch_to_numpy(y_new)



"""Module for the Reduced Order Modeling."""

import math
import copy
import pickle
import numpy as np
from scipy.spatial.qhull import Delaunay
from sklearn.model_selection import KFold

class ReducedOrderModel():
    """
    Reduced Order Model class.

    This class performs the actual reduced order model using the selected
    methods for approximation and reduction.

    :param ezyrb.Database database: the database to use for training the reduced
        order model.
    :param ezyrb.Reduction reduction: the reduction method to use in reduced order
        model.
    :param ezyrb.Approximation approximation: the approximation method to use in
        reduced order model.
    :param object scaler_red: the scaler for the reduced variables (eg. modal
        coefficients). Default is None.

    :cvar ezyrb.Database database: the database used for training the reduced
        order model.
    :cvar ezyrb.Reduction reduction: the reduction method used in reduced order
        model.
    :cvar ezyrb.Approximation approximation: the approximation method used in
        reduced order model.
    :cvar object scaler_red: the scaler for the reduced variables (eg. modal
        coefficients).

    :Example:

         >>> from ezyrb import ReducedOrderModel as ROM
         >>> from ezyrb import POD, RBF, Database
         >>> pod = POD()
         >>> rbf = RBF()
         >>> # param, snapshots and new_param are assumed to be declared
         >>> db = Database(param, snapshots)
         >>> rom = ROM(db, pod, rbf).fit()
         >>> rom.predict(new_param)

    """
    def __init__(self, database, reduction, approximation, scaler_red=None):
        self.database = database
        self.reduction = reduction
        self.approximation = approximation
        self.scaler_red = scaler_red

    def fit(self, *args, **kwargs):
        r"""
        Calculate reduced space

        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        """
        self.reduction.fit(self.database.snapshots.T)
        reduced_output = self.reduction.transform(self.database.snapshots.T).T

        if self.scaler_red:
            reduced_output = self.scaler_red.fit_transform(reduced_output)

        self.approximation.fit(
            self.database.parameters,
            reduced_output,
            *args,
            **kwargs)

        return self

    def predict(self, mu):
        """
        Calculate predicted solution for given mu
        """
        mu = np.atleast_2d(mu)
        if hasattr(self, 'database') and self.database.scaler_parameters:
            mu = self.database.scaler_parameters.transform(mu)

        predicted_red_sol = np.atleast_2d(self.approximation.predict(mu))

        if self.scaler_red:  # rescale modal coefficients
            predicted_red_sol = self.scaler_red.inverse_transform(
                predicted_red_sol)

        predicted_sol = self.reduction.inverse_transform(predicted_red_sol.T)

        if hasattr(self, 'database') and self.database.scaler_snapshots:
            predicted_sol = self.database.scaler_snapshots.inverse_transform(
                    predicted_sol.T).T

        if 1 in predicted_sol.shape:
            predicted_sol = predicted_sol.ravel()
        else:
            predicted_sol = predicted_sol.T
        return predicted_sol

    def save(self, fname, save_db=True, save_reduction=True, save_approx=True):
        """
        Save the object to `fname` using the pickle module.

        :param str fname: the name of file where the reduced order model will
            be saved.
        :param bool save_db: Flag to select if the `Database` will be saved.
        :param bool save_reduction: Flag to select if the `Reduction` will be
            saved.
        :param bool save_approx: Flag to select if the `Approximation` will be
            saved.

        Example:

        >>> from ezyrb import ReducedOrderModel as ROM
        >>> rom = ROM(...) #  Construct here the rom
        >>> rom.fit()
        >>> rom.save('ezyrb.rom')
        """
        rom_to_store = copy.copy(self)

        if not save_db:
            del rom_to_store.database
        if not save_reduction:
            del rom_to_store.reduction
        if not save_approx:
            del rom_to_store.approximation

        with open(fname, 'wb') as output:
            pickle.dump(rom_to_store, output, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(fname):
        """
        Load the object from `fname` using the pickle module.

        :return: The `ReducedOrderModel` loaded

        Example:

        >>> from ezyrb import ReducedOrderModel as ROM
        >>> rom = ROM.load('ezyrb.rom')
        >>> rom.predict(new_param)
        """
        with open(fname, 'rb') as output:
            rom = pickle.load(output)

        return rom

    def test_error(self, test, norm=np.linalg.norm):
        """
        Compute the mean norm of the relative error vectors of predicted
        test snapshots.

        :param database.Database test: the input test database.
        :param function norm: the function used to assign at the vector of
            errors a float number. It has to take as input a 'numpy.ndarray'
            and returns a float. Default value is the L2 norm.
        :return: the mean L2 norm of the relative errors of the estimated
            test snapshots.
        :rtype: numpy.ndarray
        """
        predicted_test = self.predict(test.parameters)
        return np.mean(
            norm(predicted_test - test.snapshots, axis=1) /
            norm(test.snapshots, axis=1))

    def kfold_cv_error(self, n_splits, *args, norm=np.linalg.norm, **kwargs):
        r"""
        Split the database into k consecutive folds (no shuffling by default).
        Each fold is used once as a validation while the k - 1 remaining folds
        form the training set. If `n_splits` is equal to the number of
        snapshots this function is the same as `loo_error` but the error here
        is relative and not absolute.

        :param int n_splits: number of folds. Must be at least 2.
        :param function norm: function to apply to compute the relative error
            between the true snapshot and the predicted one.
            Default value is the L2 norm.
        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        :return: the vector containing the errors corresponding to each fold.
        :rtype: numpy.ndarray
        """
        error = []
        kf = KFold(n_splits=n_splits)
        for train_index, test_index in kf.split(self.database):
            new_db = self.database[train_index]
            rom = type(self)(new_db, copy.deepcopy(self.reduction),
                             copy.deepcopy(self.approximation)).fit(
                                 *args, **kwargs)

            error.append(rom.test_error(self.database[test_index], norm))

        return np.array(error)

    def loo_error(self, *args, norm=np.linalg.norm, **kwargs):
        r"""
        Estimate the approximation error using *leave-one-out* strategy. The
        main idea is to create several reduced spaces by combining all the
        snapshots except one. The error vector is computed as the difference
        between the removed snapshot and the projection onto the properly
        reduced space. The procedure repeats for each snapshot in the database.
        The `norm` is applied on each vector of error to obtained a float
        number.

        :param function norm: the function used to assign at each vector of
            error a float number. It has to take as input a 'numpy.ndarray` and
            returns a float. Default value is the L2 norm.
        :param \*args: additional parameters to pass to the `fit` method.
        :param \**kwargs: additional parameters to pass to the `fit` method.
        :return: the vector that contains the errors estimated for all
            parametric points.
        :rtype: numpy.ndarray
        """
        error = np.zeros(len(self.database))
        db_range = list(range(len(self.database)))

        for j in db_range:
            indeces = np.array([True] * len(self.database))
            indeces[j] = False

            new_db = self.database[indeces]
            test_db = self.database[~indeces]
            rom = type(self)(new_db, copy.deepcopy(self.reduction),
                             copy.deepcopy(self.approximation)).fit(
                                 *args, **kwargs)

            error[j] = rom.test_error(test_db)

        return error

    def optimal_mu(self, error=None, k=1):
        """
        Return the parametric points where new high-fidelity solutions have to
        be computed in order to globally reduce the estimated error. These
        points are the barycentric center of the region (simplex) with higher
        error.

        :param numpy.ndarray error: the estimated error evaluated for each
            snapshot; if error array is not passed, it is computed using
            :func:`loo_error` with the default function. Default value is None.
        :param int k: the number of optimal points to return. Default value is
            1.
        :return: the optimal points
        :rtype: numpy.ndarray
        """
        if error is None:
            error = self.loo_error()

        mu = self.database.parameters
        tria = Delaunay(mu)

        error_on_simplex = np.array([
            np.sum(error[smpx]) * self._simplex_volume(mu[smpx])
            for smpx in tria.simplices
        ])

        barycentric_point = []
        for index in np.argpartition(error_on_simplex, -k)[-k:]:
            worst_tria_pts = mu[tria.simplices[index]]
            worst_tria_err = error[tria.simplices[index]]

            barycentric_point.append(
                np.average(worst_tria_pts, axis=0, weights=worst_tria_err))

        return np.asarray(barycentric_point)

    def _simplex_volume(self, vertices):
        """
        Method implementing the computation of the volume of a N dimensional
        simplex.
        Source from: `wikipedia.org/wiki/Simplex
        <https://en.wikipedia.org/wiki/Simplex>`_.

        :param numpy.ndarray simplex_vertices: Nx3 array containing the
            parameter values representing the vertices of a simplex. N is the
            dimensionality of the parameters.
        :return: N dimensional volume of the simplex.
        :rtype: float
        """
        distance = np.transpose([vertices[0] - vi for vi in vertices[1:]])
        return np.abs(
            np.linalg.det(distance) / math.factorial(vertices.shape[1]))
